package PPM::Validate;

use strict;
use warnings;

use base 'Exporter';
our @EXPORT_OK = qw(verify_file set_validator set_trace_function set_no_verification
                    get_no_verification pmx_prefix get_validator_location fetch_file);

use File::Path qw(rmtree);

# Globals (to make this work easily with PPM):

our $PPMERR;                 # Set on errors (for compatibility with PPM)
my $NO_VERIFY = 0;           # Whether or not we should verify that packages have been signed
our $VALIDATOR = '';         # The location of pmx-validator
our $TRACE_SUB = sub {};     # Code ref to log to
our $VALIDATE_VERSION   = "2";        # Version 1 didn't have a version
END {
}

sub trace {
    $TRACE_SUB->(@_);
}

# Verify that a file comes from a trusted source
sub verify_file {
    my ($file, $repo) = @_;

    if ( get_no_verification() ) {
        trace("verify_file: skipping validation of file '$file'");
        return 1;
    }

    trace("verify_file: verifying file '$file' from repository '$repo'");
    my $res = _verify(file => $file, repo => $repo);
    if ($res) {
        trace("verify_file: file '$file' verified correctly");
    } else {
        trace("verify_file: Verifying file '$file' from repository '$repo'");
        $PPM::Validate::PPMERR = "Error verifying file '$file': $PPMERR";
        trace("verify_file: $PPM::Validate::PPMERR");
    }
    return $res;
}

# Fetch a file from a trusted source
# (If verification is turned off, still download the file, just don't verify it)
sub fetch_file {
    my ($file, $repo) = @_;

    my $no_verify = get_no_verification();

    if ( $no_verify ) {
        trace("fetch_file: skipping validation of file '$file'");
        # Turn off validation so pmx-validator downloads the file anyway
    }

    trace("fetch_file: attempting to download file '$file' from repository '$repo'");
    my $res = _verify(file => $file, repo => $repo, fetch_only => 1);
    if ($res) {
        trace("fetch_file: file '$file' verified correctly");
    } else {
        trace("fetch_file: Verifying file '$file' from repository '$repo'");
        $PPM::Validate::PPMERR = "Error verifying file '$file': $PPMERR";
        trace("fetch_file: $PPM::Validate::PPMERR");
    }
    return $res;
}

# Verify a file (given as a pathname or URI)
sub _verify {
    my %args = @_;

    for my $k (qw(repo file)){
        $args{$k} = File::Spec->rel2abs($args{$k}) if -e $args{$k};
    }
    my ($repo,$file) = @args{qw(repo file)};
    my $fetch_only   = exists $args{fetch_only} ? $args{fetch_only} : 0;

    return _validator($file, $repo, $fetch_only);
}

sub _cmd_env {
    my $fetch = shift;
    my @env;
    push(@env, "VALIDATE_VER=$VALIDATE_VERSION");
    push(@env, "PMX_VALIDATOR_PPID=$$");
    push(@env, "VALIDATE_FETCH_FILE=1") if $fetch;
    return join(' ', @env);
}

sub _validator {
    my ($file, $repo, $fetch_only) = @_;
    if (!$VALIDATOR) {
        $VALIDATOR = get_validator_location();
    }

    die "Can't find package validator at '$VALIDATOR'" if (!-e $VALIDATOR);
    my $env = _cmd_env($fetch_only);
    my $output = `$env $VALIDATOR $repo $file 2>&1`;
    my $filename;
    if ( $! and ( ($filename) = $output =~ /^(.+): OK\s?$/m ) ) {
        trace($output) if $ENV{DEBUG_VALIDATED_FETCH};
        return $filename;
    } else {
        $PPMERR = $output;
        return '';
    }
}

sub get_validator_location {
    my $PREFIX; # We do a runtime import of $PREFIX, 
                # so this is to allow the module to compile
    eval {
        require PureMessage::Version;
        $PREFIX = $PureMessage::Version::PREFIX;
    };
   
    if ($@) {
        # Try finding it anyway
        $PREFIX = pmx_prefix();         
    } 
    # If you got this message, you need to set $VALIDATOR yourself
    die "Could not locate pmx-validator" if !$PREFIX;
    
    return "$PREFIX/bin/pmx-validator";
}

sub set_validator {
    $VALIDATOR = shift;
}

sub set_trace_function {
    $TRACE_SUB = shift;
}

sub set_no_verification {
    my $no = shift;
    $NO_VERIFY = $no ? 1 : 0;
}

sub get_no_verification {
    return 1 if -e '/etc/puremessage-novalidation' || $ENV{PMX_NO_VALIDATE_PACKAGES};
    return $NO_VERIFY;
}

sub pmx_prefix {
  my $PREFIX;
  if(-e "/etc/puremessage-5") {
    open my $fh, "/etc/puremessage-5";
    my $line;
    # There can be multiple lines in the prefix file, so we
    # take the last one (which will have been most recently added)
    while (<$fh>) {
        $line = $_ if ($_ =~ m{^/});
    }
    if ($line) {
        chomp $line;
        $PREFIX = $line;
    }
    close $fh;
  }

  return $PREFIX;
}

1;

