package Sophos::Config;

use strict;
use XSLoader ();
use base 'Exporter';

our $VERSION = '2.100';

# The XS layer pushes a few constants onto @EXPORT_OK, so set up %EXPORT_TAGS
# after doing that.
our @EXPORT_OK = (qw(
    path_nodes
    path_node
    path_value
    path_values
    path_value_append
    path_children
));
XSLoader::load(__PACKAGE__, $VERSION);
our %EXPORT_TAGS = (
    path  => [grep /^path_/,  @EXPORT_OK],
    visit => [grep /^VISITOR_/, @EXPORT_OK],
);

sub new {
    my $class = shift;
    my %opts  = (
	die_on_error  => 1,
	auto_validate => 1,
	strict        => 0,
	@_,
    );

    my $self = _alloc_cf(\%opts);
    my $priv = $self->appdata;

    $priv->{_die_on_error}  = delete $opts{die_on_error};
    $priv->{_auto_validate} = delete $opts{auto_validate};
    $priv->{strict}         = delete $opts{strict};
    $priv->{default_type}   = delete $opts{default_type};
    $priv->{on_save}        = delete $opts{on_save};
    $priv->{on_delete}      = delete $opts{on_delete};

    # Provide default arguments to $self->parse_file()
    my $pf = $priv->{pf_opts} = {};
    for my $key (qw(backup_ext rotate timeout)) {
	if (exists $opts{$key}) {
	    $pf->{$key} = delete $opts{$key};
	}
    }

    return $self;
}

#=============================================================================
# The Parser API
#=============================================================================

sub parse_file; # implemented in XS
sub parse;      # implemented in XS
sub root;       # implemented in XS
sub as_string;  # implemented in XS
sub saveto;     # implemented in XS
sub save;       # implemented in XS
sub as_hashref; # implemented in XS
sub as_arrayref; # implemented in XS

sub as_hash {
    my $self = shift;
    my $href = $self->as_hashref;
    wantarray ? %$href : $href;
}

sub as_list {
    my $self = shift;
    my $aref = $self->as_arrayref;
    wantarray ? @$aref : $aref;
}

sub unlock;     # implemented in XS
sub node;       # implemented in XS

sub count_parse_errors; # implemented in XS
sub get_parse_error;    # implemented in XS

sub count_validator_errors {
    my $self = shift;
    my $data = $self->appdata;
    my $errs = $data->{vl}{errors};
    return $errs ? @$errs : 0;
}

sub get_validator_error {
    my $self = shift;
    my $index = shift;
    my $data = $self->appdata;
    my $errs = $data->{vl}{errors};
    return $errs ? $errs->[$index] : undef;
}

sub modified {
    my $self = shift;
    my $mod = 0;
    $self->root->visit(sub {
	my $ref = shift;
	my $node = shift;
	++$$ref if $node->modified;
	return VISITOR_CONTINUE();
    }, \$mod);
    return $mod;
}

#=============================================================================
# Validators
#=============================================================================

# Adds the {vl} slot to $self, where all validators, formatters, typedefs, and
# path declarations are stored and manipulated. This is created lazily,
# whenever the application calls a method the implies validation must occur.
sub alloc_vl {
    my $self = shift;
    my $priv = $self->appdata;
    require Sophos::Config::Validator;
    $priv->{vl} = Sophos::Config::Validator->new($self);
}

# If $priv->{vl} is defined, then $self->validate() is called by parse() and
# parse_file() -- unless `auto_validate => 0' was passed to new(). The entire
# tree is validated by loading all 
sub validate {
    my $self = shift;
    my $priv = $self->appdata;
    alloc_vl($self) unless $priv->{vl};
    Sophos::Config::Validator::validate($self);
}

sub format {
    my $self = shift;
    my $priv = $self->appdata;
    alloc_vl($self) unless $priv->{vl};
    Sophos::Config::Validator::format($self);
}

# Version 1.0 typedecl() function.
sub typedecl {
    my $self = shift;
    my $priv = $self->appdata;
    alloc_vl($self) unless $priv->{vl};

    # Define a new typedef
    if (@_ and $_[0] =~ /^#/) {
	return $priv->{vl}->typedef(@_);
    }
    # Retrieve validator and formatter for a typedef
    elsif (@_ == 1 and $_[0] =~ /^#/) {
	return $priv->{vl}->typedef($_[0]);
    }
    # Retrieve or declare a path declaration
    else {
	return $priv->{vl}->declare(@_);
    }
}

# Version 2.0 typedefs. 
sub typedef {
    my $self = shift;
    my $priv = $self->appdata;
    alloc_vl($self) unless $priv->{vl};
    return $priv->{vl}->typedef(@_);
}

sub declare {
    my $self = shift;
    my $priv = $self->appdata;
    alloc_vl($self) unless $priv->{vl};
    return $priv->{vl}->declare(@_);
}

sub default {
    my $self = shift;
    my $priv = $self->appdata;
    alloc_vl($self) unless $priv->{vl};
    return $priv->{vl}->default(@_);
}

#=============================================================================
# The Path API
#=============================================================================

sub path_nodes {
    my ($root, $path, %opts) = @_;

    # These are probably not even useful anymore, but we do support them.
    if (my $insert = $opts{insert}) {
	return $root->attach($path, $insert) ? $insert : undef;
    }
    elsif ($opts{delete}) {
	return $root->delete($path);
    }

    # Select (or create) nodes in the tree.
    my $create = delete $opts{create} || 0;
    my $cursor = $root->select($path, $create) or return;

    # Call 'vivify' if autovalidation is enabled
    if ($create) {
	my $parser = $root->parser;
	my $data = $parser->appdata;
	if ($data->{vl} && $data->{_auto_validate}) {
	    Sophos::Config::Validator::vivify($data->{vl}, $parser);
	}
    }

    return $cursor unless wantarray;

    my @matched;
    while (my $n = $cursor->node) {
	$cursor->next;
	push @matched, $n;
    }
    return @matched;
}

sub path_node {
    my $ret = path_nodes(@_) or return undef;

    # If it was a cursor, return the first node.
    return $ret->node if UNIVERSAL::can($ret, "node");

    # Otherwise just return it.
    return $ret;
}

sub path_value {
    my ($root, $path, @value) = @_;
    my $create = @_ >= 3;
    my $node = path_node($root, $path, create => $create)
	or return;
    if ($create) {
	$node->value(\@value);
	return;
    }
    else {
	# XXX shouldn't be needed -- $node->value should flatten in
	# array context.
	return $node->values if wantarray;
	return $node->value;
    }
}

sub path_values {
    my ($root, $path, @values) = @_;
    my $node = path_node($root, $path, create => @values ? 1 : 0) or return;
    return $node->values unless @values; # set the values
    return $node->value(\@values);
}

sub path_value_append {
    my ($root, $path) = splice(@_, 0, 2);
    my $node = path_node($root, $path, create => 1);
    $node->value_append($_) for @_;
}

#=============================================================================
# The Node API
#=============================================================================
package Sophos::Config::Node;

sub seen;           # implemented in XS
sub modified;       # implemented in XS
sub hint;           # implemented in XS
sub kind;           # implemented in XS
sub path;           # implemented in XS
sub name;           # implemented in XS
sub group;          # implemented in XS
sub comment;        # implemented in XS
sub is_list;        # implemented in XS
sub value;          # implemented in XS
sub value_append;   # implemented in XS
sub detach;         # implemented in XS
sub as_string;      # implemented in XS
sub as_hashref;     # implemented in XS
sub set_to_default; # implemented in XS

# Insert nodes in places.
sub insert;         # implemented in XS
sub insertafter;    # implemented in XS
sub append;         # implemented in XS

# Use path_*() functions without importing them.
sub select;         # implemented in XS
sub attach;         # implemented in XS
sub delete;         # implemented in XS

# Tree traversal
sub next;           # implemented in XS
sub prev;           # implemented in XS
sub children;       # implemented in XS
sub parent;         # implemented in XS

# Save this node and all its children to "file.conf" (which must have already
# been !included in the original parse)
sub path_filename;  # implemented in XS

# Backwards-compatible wrappers
*dirty          = \&modified;
*groupclass     = \&group;
*block_comment  = \&comment;
*_as_string     = \&as_string;

sub typedecl {
    warn "\$node->typedecl() is deprecated: use \$node->declared() instead.\n";
    return declared(@_);
}

sub tokenizer {
    warn "\$node->tokenizer() is deprecated: use \$node->io() instead.\n";
    return io(@_);
}

sub type {
    warn "\$node->type() is deprecated: use \$node->kind() instead.\n";
    return kind(@_);
}

sub values {
    my $self = shift;
    return $self->value([@_]) if @_;
    if ($self->is_list) {
	return $self->count_values unless wantarray;
	my $v = $self->value;
	return @$v;
    }
    return $self->value;
}

# This one was always *really* messy. I wish we could do something better.
sub line_comment {
}

# If $parser->{vl} is defined, then $node->format() is called by any method
# that modifies a node's value -- unless `auto_validate => 0' was passed to
# new().
sub format {
    my $self = shift;
    my $parser = shift;
    Sophos::Config::Validator::_format_node($parser, $self);
}
sub validate {
    my $self = shift;
    my $parser = shift;
    Sophos::Config::Validator::_validate_node($parser, $self);
}

sub as_hash {
    my $self = shift;
    my $href = $self->as_hashref;
    wantarray ? %$href : $href;
}

package Sophos::Config::Cursor;

*size = \&more;

1;
