package Sophos::Config::Validator;

use strict;
use Sophos::Config qw(:visit);
use Data::Dump 'dump';

sub new {
    my ($klass, $parser) = @_;
    my $priv = $parser->appdata;
    my $self = bless {
	type => {},
	decl => {},
	dflt => {},
	default_type => $priv->{default_type},
    }, ref($klass) || $klass;
    register_types($self);
    return $self;
}

sub typedef {
    my ($self, $type, $class) = @_;
    $self->{type}{$type} = $class if defined $class;
    return $self->{type}{$type};
}

sub declare {
    my ($self, $path, $type) = @_;
    if (defined $type) {
	if (defined $self->{default_type}) {
	    my ($prepend, @exceptions) = @{$self->{default_type}};
	    my $matched = 0;
	    for my $exc (@exceptions) {
		if ($type =~ /^\Q$exc\E/) {
		    $matched = 1;
		    last;
		}
	    }
	    $type = "$prepend $type" unless $matched;
	}
	$self->{decl}{$path} = $type;
    }
    return $self->{decl}{$path};
}

sub default {
    my ($self, $path, $default) = @_;
    $self->{dflt}{$path} = $default if defined $default;
    return $self->{dflt}{$path};
}

#=============================================================================
# Validator functions
#=============================================================================

sub vivify {
    my ($self, $parser) = @_;
    my $root = $parser->root;
    for my $pattern (keys %{$self->{dflt}}) {
	my $dflt = $self->{dflt}{$pattern};
	next unless defined $dflt;
	my $ref = ref($dflt);

	my $get_value = sub {
	    return $dflt if $ref eq '';
	    return $dflt if $ref eq 'ARRAY';

	    my $node = shift;
	    return scalar &$dflt($node) if $ref eq 'CODE';

	    my $path = $node->path;
	    die "Illegal default value $dflt at '$path'.\n";
	};

	# select nodes which were seen in the tree.
	if (defined(my $cursor = $root->select($pattern))) {
	    while (my $node = $cursor->node) {
		$node->_set_default(&$get_value($node));
		$cursor = $cursor->next;
	    }
	}

	# vivify nodes that haven't been created.
	if (defined(my $cursor = $parser->_vivify_defaults($pattern))) {
	    while (my $node = $cursor->node) {
		$node->_set_default(&$get_value($node));
		$cursor = $cursor->next;
	    }
	}
    }
}

# These are functions, not methods. The parser object is the only parameter.
# The validator is $parser->appdata->{vl}.
sub validate {
    my $parser = shift;
    my $node   = $parser->root;

    my $data   = $parser->appdata;
    delete $data->{vl}{errors};

    # First validate pre-default values
    $node->visit(\&_prevalidate_node, $parser);

    # Now create all the defaults
    vivify($data->{vl}, $parser);

    # Now validate post-default values
    $node->visit(\&_validate_node, $parser);

    # If we're supposed to die on error, throw the first error.
    if ($data->{vl}{errors} && $data->{_die_on_error}) {
	die $data->{vl}{errors}[0]{message};
    }

    return;
}

sub format {
    my $parser = shift;
    my $node   = $parser->root;

    my $data   = $parser->appdata;
    delete $data->{vl}{errors};

    $node->visit(\&_format_node, $parser);

    # If we're supposed to die on error, throw the first error.
    if ($data->{vl}{errors} && $data->{_die_on_error}) {
	die $data->{vl}{errors}[0]{message};
    }

    return;
}

sub _prevalidate_node {
    my $parser = shift;
    my $node   = shift;
    my $data = $parser->appdata;

    # Look up this node in the tables
    my $self = $data->{vl};
    my $path = $node->path;
    my $decl = lookup_decl($self, $path, 0);

    return VISITOR_PRUNE if $decl eq '#any';
    execute_decl($decl, $self, $node, $path, 'prevalidate');

    return VISITOR_CONTINUE;
}

# Ensure that $node's formatted value is within the range of acceptable values
# for the node's type. If the node has no declared type, this always succeeds.
# If the check succeeds, the value returned from the validator is stored in
# the native type.
sub _validate_node {
    my $parser = shift;
    my $node   = shift;
    my $data = $parser->appdata;

    # Look up this node in the tables
    my $self = $data->{vl};
    my $path = $node->path;
    my $decl = lookup_decl($self, $path, 1);

    if ($decl) {
	return VISITOR_PRUNE if $decl eq '#any';
	execute_decl($decl, $self, $node, $path, 'validate');
    }
    elsif ($data->{strict} and $node->kind eq 'option') {
	push @{$self->{errors}}, _new_error($node, undef, [],
	    "unexpected option '$path' while parsing in strict mode\n");
    }

    return VISITOR_CONTINUE;
}

sub _format_node {
    my $parser = shift;
    my $node   = shift;

    # Look up this node in the tables
    my $self = $parser->appdata->{vl};
    my $path = $node->path;
    my $decl = lookup_decl($self, $path, 0);

    return VISITOR_PRUNE if $decl eq '#any';
    execute_decl($decl, $self, $node, $path, 'format');

    return VISITOR_CONTINUE;
}

sub execute_decl {
    my $decl = shift;
    my $self = shift;
    my $node = shift;
    my $path = shift;
    my $what = shift;

#print "execute_decl('$decl', '$path', '$what')\n";

    my @part = split ' ', $decl;
    while (@part) {
	for my $type (shift(@part), '#str') {
	    my @args;
	    push @args, shift @part while @part && substr($part[0],0,1) ne '#';
	    my $class = $self->{type}{$type}
		or die "unknown type '$type' in '$decl'\n";

	    my $meth;
	    if (defined ($meth = UNIVERSAL::can($class, "${what}_node"))) {
		eval { $class->$meth($node, \@args) };
		if ($@) {
		    push @{$self->{errors}}, _new_error($node, $type, \@args, $@);
		    $@ = '';
		}
	    }
	    elsif (defined ($meth = UNIVERSAL::can($class, "${what}_value"))) {
		my $value = $node->Value;
		while ($value) {

		    # validate_value() should only be called on "normal"
		    # values. Format, on the other hand, should be called for
		    # any kind of node that needs formatting.
		    my $kind = $value->kind;
		    last if $what eq 'validate' and $kind ne 'normal';

		    eval { $class->$meth($node, $value, \@args) };
		    if ($@) {
			push @{$self->{errors}}, _new_error($node, $type,
							    \@args, $@);
			$@ = '';
		    }
		    $value = $value->next;
		}
	    }
	    else {
		# no $what method in this class; try the #str validator
		next;
	    }
	    # found a $what method -- skip the #str validator
	    last;
	}
    }
}

sub lookup_decl {
    my ($self, $path, $strict) = @_;

    # Try for the simple case
    if ($self->{decl}{$path}) {
	return $self->{decl}{$path};
    }

    # Now do wildcards.
    for my $pat (keys %{$self->{decl}}) {
	if (Sophos::Config::cf_path_match($pat, $path)) {
	    return $self->{decl}{$pat};
	}
    }

    return undef if $strict;

    # Return '#str'
    return $self->{default_type} ? "$self->{default_type}[0] #str" : "#str";
}

sub _new_error {
    my ($node, $type, $args, $error) = @_;
    my $path = $node->path;

    # Ensure no further formatting happens on the error message
    chomp $error;
    $error = "$path: $error\n";

    return {
	type => $type,
	args => $args,
	path => $path,
	message => $error,

# XXX it really sucks, but I can't hold onto the node object here, because it
# holds an internal reference to the parser, so we'll end up not freeing
# anything until perl's global cleanup stage. One option is to write a
# "Sophos::Config::FakeNode" class which grabs all the information out of the
# node and stuffs it into a Perl hash at the time the error is thrown.
#	node => $node,
    };
}

sub register_types {
    my $self = shift;
    for (qw(bool int num percentage duration enum str list atom readonly any))
    {
	$self->typedef("#$_", "Sophos::Config::Validator::$_");
    }
}

package Sophos::Config::Validator::bool;

sub validate_value {
    my ($self, $node, $value, $args) = @_;
    my $h = $value->appdata;
    my $v = $value->_get_formatted;
    my $native;
    if ($v eq 'yes' or $v eq '1') {
	$h->{style} = 'yesno';
	$value->_set_native(1);
    }
    elsif ($v eq 'no' or $v eq '0') {
	$h->{style} = 'yesno';
	$value->_set_native(0);
    }
    elsif ($v eq 'on') {
	$h->{style} = 'onoff';
	$value->_set_native(1);
    }
    elsif ($v eq 'off') {
	$h->{style} = 'onoff';
	$value->_set_native(0);
    }
    elsif ($v eq 'true') {
	$h->{style} = 'truefalse';
	$value->_set_native(1);
    }
    elsif ($v eq 'false') {
	$h->{style} = 'truefalse';
	$value->_set_native(0);
    }
    else {
	die "invalid boolean value '$v'\n";
    }
}

sub format_value {
    my ($self, $node, $value, $args) = @_;
    my $h = $value->appdata;
    my $style = $h->{style} || 'yesno';
    defined(my $v = $value->_get_native) or return;
    die "invalid boolean value '$v'\n" unless $v eq '0' or $v eq '1';
    if ($style eq 'yesno') {
	$value->_set_formatted($v ? 'yes' : 'no');
    }
    elsif ($style eq 'onoff') {
	$value->_set_formatted($v ? 'on' : 'off');
    }
    elsif ($style eq 'truefalse') {
	$value->_set_formatted($v ? 'true' : 'false');
    }
}

package Sophos::Config::Validator::int;

my $dec_rx = qr/([-+]?)([0-9]+)/;
my $oct_rx = qr/([-+]?)(0[0-7]+)/;
my $hex_rx = qr/([-+]?)(0[Xx][0-9A-Fa-f]+)/;
my %sign   = ('-' => -1,
	      '+' => 1,
	      ''  => 1);

sub validate_value {
    my ($self, $node, $value, $args) = @_;
    my $h = $value->appdata;
    my $v = $value->_get_formatted;
    if ($v =~ /^$hex_rx$/) {
	my ($sign, $mantissa) = ($1, $2);
	$h->{style} = 'hex';
	$value->_set_native($sign{$sign} * hex($mantissa));
    }
    elsif ($v =~ /^$oct_rx$/) {
	my ($sign, $mantissa) = ($1, $2);
	$h->{style} = 'oct';
	$value->_set_native($sign{$sign} * oct($mantissa));
    }
    elsif ($v =~ /^$dec_rx$/) {
	my ($sign, $mantissa) = ($1, $2);
	$h->{style} = 'dec';
	$value->_set_native($sign{$sign} * $mantissa);
    }
    else { die "invalid integer '$v'\n" }
}

sub format_value {
    my ($self, $node, $value, $args) = @_;
    my $h = $value->appdata;
    defined(my $v = $value->_get_native) or return;
    die "invalid integer '$v'\n" unless $v =~ /^-?\d+$/;
    my $neg = '';
    my $style = $h->{style} || 'dec';
    if ($v < 0) { $v = abs($v); $neg = '-' }
    if ($style eq 'hex') {
	$value->_set_formatted(sprintf("${neg}0x%x", $v));
    }
    elsif ($style eq 'oct') {
	$value->_set_formatted(sprintf("${neg}0%o", $v));
    }
    else {
	$value->_set_formatted(sprintf("${neg}%d", $v));
    }
}

package Sophos::Config::Validator::num;

my $float_rx = qr/(?:(?i)(?:[+-]?)(?:(?=[0123456789]|[.])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[E])(?:(?:[+-]?)(?:[0123456789]+))|))/;

sub validate_value {
    my ($self, $node, $value, $args) = @_;
    my $v = $value->_get_formatted;
    die "invalid number '$v'\n" unless $v =~ /^$float_rx$/;
}

sub format_value {
    my ($self, $node, $value, $args) = @_;
    defined(my $v = $value->_get_native) or return;
    die "invalid number '$v'\n" unless $v =~ /^$float_rx$/;
}

package Sophos::Config::Validator::percentage;

sub validate_value {
    my ($self, $node, $value, $args) = @_;
    my $v = $value->_get_formatted;
    die "invalid percentage '$v'\n" unless $v =~ s/^($float_rx)(%?)$/$1/;
    if ($2) {
	$v /= 100;
	$value->_set_native($v);
	$value->appdata->{percent} = 1;
    }
}

sub format_value {
    my ($self, $node, $value, $args) = @_;
    defined(my $v = $value->_get_native) or return;
    die "invalid percentage '$v'\n" unless $v =~ /^$float_rx$/;
    my $h = $value->appdata;
    if ($h->{percent} or $v >= 0 and $v <= 1) {
	$v *= 100;
	$v .= "%";
    }
    $value->_set_formatted($v);
}

package Sophos::Config::Validator::duration;

sub validate_value {
    my ($self, $node, $value, $args) = @_;
    my $v = $value->_get_formatted;
    my $tot = 0;
    my @parts = grep length, split m/($float_rx(?:)[smhdwy]?)/, $v;
    for my $part (@parts) {
	my ($n, $t) = ($part =~ /($float_rx)([smhdwy]?)/);
	die "invalid duration '$v'\n" unless defined $n;
	if    ($t eq 's' or !$t) { $tot += $n }
	elsif ($t eq 'm')        { $tot += $n * 60 }
	elsif ($t eq 'h')        { $tot += $n * 3600 }
	elsif ($t eq 'd')        { $tot += $n * 86400 }
	elsif ($t eq 'w')        { $tot += $n * 604800 }
	elsif ($t eq 'y')        { $tot += $n * 31536000 }
	else {
	    die "invalid duration '$v'\n";
	}
    }
    $value->_set_native($tot);
}

sub format_value {
    my ($self, $node, $value, $args) = @_;
    defined(my $v = $value->_get_native) or return;
    unless ($v) {
	$value->_set_formatted("0s");
	return;
    }
    die "invalid duration '$v'\n" unless $v =~ /^$float_rx$/;

    my $neg = $v =~ s/^-// ? '-' : '';
    my %t = map { $_ => 0 } qw(y w d h m s);
    my %v = (
	y => 31536000,
	w => 604800,
	d => 86400,
	h => 3600,
	m => 60,
    );
    for my $fac (qw(y w d h m)) {
	my $val = $v;
	while ($val >= 0) {
	    $val -= $v{$fac};
	    $t{$fac}++;
	}
	$t{$fac}--;
	$v = $val + $v{$fac};
	last unless $v;
    }
    $t{s} = $v;
    my $f = $neg . join '', map "$t{$_}$_", grep $t{$_}, qw(y w d h m s);
    $value->_set_formatted($f);
}

package Sophos::Config::Validator::enum;

sub validate_value {
    my ($self, $node, $value, $args) = @_;
    my $v = $value->_get_formatted;
    for my $ok (@$args) {
	if ($v eq $ok) {
	    return; # okay
	}
    }
    die "invalid value '$v': should be one of '@$args'\n";
}

sub format_value {
    my ($self, $node, $value, $args) = @_;
    defined(my $v = $value->_get_native) or return;
    for my $ok (@$args) {
	if ($v eq $ok) {
	    $value->_set_formatted($v);
	    return; # okay
	}
    }
    die "invalid value '$v': should be one of '@$args'\n";
}

package Sophos::Config::Validator::list;

sub validate_node {
    my ($self, $node, $args) = @_;
    $node->appdata->{is_list} = 1;
    $node->kind("option");
}

sub format_node { } # do nothing

package Sophos::Config::Validator::atom;

sub validate_node {
    my ($self, $node, $args) = @_;
    if ($node->is_list) {
	my $mode = shift(@$args) || 'over';
	if ($mode eq 'over') {
	    my $v = $node->Value;
	    while ($v->next) { $v = $v->next }
	    $v->detach;
	    $node->Value($v);
	}
	elsif ($mode eq 'nodups') {
	    my $count = $node->count_values;
	    die "node is a list of $count, but is declared as an atom.\n";
	}
	elsif ($mode eq 'first') {
	    my $v = $node->Value;
	    while ($v->next) {
		$v->next->detach;
	    }
	}
    }
}

sub format_node { } # do nothing

package Sophos::Config::Validator::readonly;

sub validate_node {
    my ($self, $node, $args) = @_;
    my $path = $node->path;
    die "can't set readonly node $path\n" if $node->seen;
}

sub format_node { } # do nothing

package Sophos::Config::Validator::any;

package Sophos::Config::Validator::str;

sub format_value {
    my ($self, $node, $value, $args) = @_;
    defined(my $v = $value->_get_native) or return;
    $value->_set_formatted($v);
}

1;

__END__

=head1 NAME

Sophos::Config::Validator - support module for config validation

=head1 SYNOPSIS

none

=head1 DESCRIPTION

A validating Sophos::Config parser contains three node trees:

=over

=item 1

The I<raw> tree. Nodes and values are exactly as they appeared in the config
file.

=item 2

The I<defaults> tree. Nodes and values are created based on the defaults
specified by the application, and what nodes the I<raw> tree contains. If a
node exists in the I<raw> tree, it is I<not> present in the I<defaults> tree.

=item 3

The I<native> tree. Nodes and values are in their "native" format, which
may be quite different from their I<raw> format. In particular, the node could
be associated with a C<typedef> which defines a custom C<load()> and C<dump()>
method to translate between I<native> and I<raw>.

=back

=head2 Built In Types

Sophos::Config::Validator comes with these types built in:

=over 4

=item #bool

Requires the value to be one of the values C<true>, C<false>, C<yes>, C<no>,
C<on>, C<off>, C<1>, or C<0>. Internally, the value is always C<0> or C<1>,
but this also attaches a formatter that outputs the value the same way it was
read in, even if the value is reversed. For example, if the values was
originally "yes" in the file, but the setting was disabled, the value would be
written out as "no".

If there are no semantic hints in the file, the formatter will choose "true"
and "false" as the preferred way to write it. You can override this by setting
the node to a particular value like "yes".

=item #int

Requires the value to be an integer. Octal and hex formats are supported. The
formatter writes the value out in the same format as it was read in. To
override this, you can set the value to a particular format, like "0x25".

=item #num

Requires the value to be any decimal number. The formatter is a no-op, which
means Perl's own stringification is used.

=item #percentage

Requires the value to be a #num, possibly followed by a C<%>. The C<%>
means the value should be divided by 100. There is no checking for numbers
outside the range of 0-100%.

When writing out, numbers between 0 and 1 are formatted with a C<%> sign, even
if they were read as normal numbers. Other numbers are output as either plain
numbers or with a C<%> sign, depending on how they were read.

=item #duration

Reads a sendmail-style duration and converts it to seconds. Adds an
output formatter to save the value in the same format.

For example:

   5m30s        # 330
   5.5m         # 330
   5.5h         # 19800
   22.1s        # 22.1
   2y4w1d3h2m   # 65588520

Always writes the value out so as to favour the largest units possible, while
keeping integer coefficients. So the value 1452.7 would always be written:

   24m12.7s

If a number does not have a unit, it is assumed to be seconds, and is not
converted.

=item #enum

An enumerated type. You must provide the names allowed. In the example below,
the values "foo", "bar", and "baz" are valid; any other value will throw an
exception:

   $p->typedecl("path", "#enum foo bar baz");

The formatter is a no-op: values are just kept as strings.

=item #str

Does no validation or formatting. Everything is a string.

=item #list

Forces listification of the node. No changes are made to the value, and no
formatter is added. You can specify the types of each list element:

   $p->typedecl("foo.bar", "#list #int");
   $p->typedecl("foo.bar", "#list #enum foo bar");

If you do not specify an element, "#str" is assumed.

=item #atom

Forces non-listification of the node. No changes are made to the value, and no
formatter is added. You can specify whether repeated instances of this node in
the config file will overwrite the first value, or cause an error:

   $p->typedecl('repeats-allowed', '#atom');
   $p->typedecl('repeats-okay-too', '#atom over');
   $p->typedecl('no-repeats', '#atom nodups');

=item #readonly

Forces the value to be "readonly", and causes an error if the value is seen in
the config file. This is useful if you want to ensure that a value exists in
the %config hash (from as_hash()) but you do not want users to override it. In
that case, you can simply set the value with default(), and lock it with the
#readonly type.

=item #any

Disables strictness for this node and any children. Does no validation or
formatting; this is mostly useful when you want to be strict for I<most>
nodes, but you don't care about a particular localized tree. The nodes will be
parsed as if strict mode was not enabled. You can still use validator for a
child of this node, but if you do, then I<its> children will have strictness
re-enabled. For example, if you define these:

   $p->typedecl('no-strict', '#any');
   $p->typedecl('no-strict.integer', '#int');
   $p->typedecl('no-strict.some-group', '#some-validator');

Then this will throw an error on the C<no-strict.some-group.foo> node:

   <no-strict>
       integer = 10
       <some-group>
           foo = bar
       </some-group>
   </no-strict>

That's because the parser only looks at the immediate parent of each node just
before it throws a "strict" error. If you redefine a child of an "#any" node,
it will no longer be #any.

=back

